import { Archetype } from "./Archetype";
import { commandQueue } from "./util/commandQueue";
import { idGenerator } from "./util/idGenerator";
import normalizeArchetype from "./util/normalizeArchetype";
export class World {
    /** An array holding all entities known to this world. */
    entities = new Array();
    /** An ID generator we use for assigning IDs to newly added entities. */
    nextId = idGenerator(1);
    /** A list of known archetypes. */
    archetypes = new Map();
    createArchetype(...query) {
        const normalizedQuery = normalizeArchetype(typeof query[0] === "string"
            ? { all: query }
            : query[0]);
        /* We may already have an archetype representing the same query */
        const stringifiedQuery = JSON.stringify(normalizedQuery);
        if (this.archetypes.has(stringifiedQuery))
            return this.archetypes.get(stringifiedQuery);
        /* Once we reach this point, we need to create a new archetype. */
        const archetype = new Archetype(normalizedQuery);
        this.archetypes.set(stringifiedQuery, archetype);
        for (const entity of this.entities)
            archetype.indexEntity(entity);
        return archetype;
    }
    indexEntity(entity) {
        for (const archetype of this.archetypes.values())
            archetype.indexEntity(entity);
    }
    /* MUTATION FUNCTIONS */
    addEntity = (entity) => {
        /* If there already is an ID, raise an error. */
        if ("id" in entity)
            throw "Attempted to add an entity that aleady had an 'id' component.";
        /* Assign an ID */
        entity.id = this.nextId();
        /* Store the entity... */
        this.entities.push(entity);
        /* ...and add it to relevant indices. */
        this.indexEntity(entity);
        return entity;
    };
    removeEntity = (entity) => {
        /* Remove entity from all archetypes */
        for (const archetype of this.archetypes.values())
            archetype.removeEntity(entity);
        /* Remove its id component */
        delete entity.id;
        /* Remove it from our global list of entities */
        const pos = this.entities.indexOf(entity, 0);
        this.entities.splice(pos, 1);
    };
    addComponent = (entity, name, data) => {
        entity[name] = data;
        this.indexEntity(entity);
    };
    removeComponent = (entity, ...components) => {
        components.forEach((name) => delete entity[name]);
        this.indexEntity(entity);
    };
    /* QUEUED MUTATION FUNCTIONS */
    /**
     * A queue of commands to run.
     */
    queuedCommands = commandQueue();
    queue = {
        addEntity: (entity) => {
            this.queuedCommands.add(() => this.addEntity(entity));
            return entity;
        },
        removeEntity: (entity) => {
            this.queuedCommands.add(() => this.removeEntity(entity));
        },
        addComponent: (entity, name, data) => {
            this.queuedCommands.add(() => this.addComponent(entity, name, data));
        },
        removeComponent: (entity, ...names) => {
            this.queuedCommands.add(() => this.removeComponent(entity, ...names));
        },
        flush: () => {
            this.queuedCommands.flush();
        }
    };
    /**
     * Clear the entire world by discarding all entities and indices.
     */
    clear() {
        /* Remove all entities */
        this.entities.length = 0;
        /* Remove all archetype indices */
        this.archetypes.clear();
        /* Clear queue */
        this.queuedCommands.clear();
    }
}
