import { jsx as _jsx } from "react/jsx-runtime";
import { createContext, useContext, useEffect, useState } from "react";
import { useRerender } from "./util/useRerender";
/**
 * Create various React-specific hooks and components for your
 * Miniplex ECS instance.
 *
 * @param world An instance of a Miniplex ECS to use.
 */
export function createReactIntegration(world) {
    const EntityContext = createContext(null);
    /**
     * A React component to either create a new entity, or represent an existing entity so
     * it can be enhanced with additional components (see the <Component> component.)
     */
    const Entity = ({ entity: existingEntity, children }) => {
        /* Reuse the specified entity or create a new one */
        const [entity] = useState(() => existingEntity ?? {});
        /* If the entity was freshly created, manage its presence in the ECS world. */
        useEffect(() => {
            if (existingEntity)
                return;
            world.addEntity(entity);
            return () => world.removeEntity(entity);
        }, [entity]);
        /* Provide a context with the entity so <Component> components can be wired up. */
        return _jsx(EntityContext.Provider, { value: entity, children: children }, void 0);
    };
    /**
     * Return the current entity context.
     */
    function useEntity() {
        return useContext(EntityContext);
    }
    /**
     * Declarative declare a component on an entity.
     */
    function Component({ name, data }) {
        const entity = useEntity();
        useEffect(() => {
            world.addComponent(entity, name, data);
            return () => world.removeComponent(entity, name);
        }, [entity, name, data]);
        return null;
    }
    /**
     * Return the entities of the specified archetype and subscribe this component
     * to it, making it re-render when entities are added to or removed from it.
     */
    function useArchetype(...query) {
        const rerender = useRerender();
        const archetype = world.createArchetype(...query);
        useEffect(() => {
            archetype.onEntityAdded.on(rerender);
            archetype.onEntityRemoved.on(rerender);
            return () => {
                archetype.onEntityAdded.off(rerender);
                archetype.onEntityRemoved.off(rerender);
            };
        }, [world, archetype]);
        return archetype.entities;
    }
    return { useArchetype, useEntity, Entity, Component };
}
